# isf_reader.py

# Code by Evgeniia Martuganova, TU Delft, 2022, using Jupyter notebooks.
# Notebook adapted to function by Tom Kettlety, UoOxford, 2023.

import re
import numpy as np
import pandas as pd

def read_origin_block_line(line):
    # Origin_block_dict = {'Origin_Date': [], 'Origin_Time': [], 'Origin_time_flag': [],\
    # 'Origin_time_error': [],'Origin_RMS': [],'Origin_lat': [],'Origin_long': [],\
    # 'Origin_epicenter_flag': [], 'Origin_Smaj': [], 'Origin_Smin': [], 'Origin_Az' : [], \
    # 'Origin_Depth': [], 'Depth_fixed_flag': [], 'Origin_depth_errors': [], \
    # 'Origin_Ndef': [],'Origin_Nsta': [],'Origin_Gap': [],'Origin_mdist': [],\
    # 'Origin_Mdist': [],'Origin_analysis_type': [], 'Origin_location_method': [], \
    # 'Origin_event_type': [], 'Origin_author': [],'Origin_ID': []}

    Origin_block_dict['Origin_Date'].extend([line[0:10]])
    Origin_block_dict['Origin_Time'].extend([line[11:22]])
    Origin_block_dict['Origin_time_flag'].extend([line[22:23]])
    Origin_block_dict['Origin_time_error'].extend([line[24:29]])
    Origin_block_dict['Origin_RMS'].extend([line[30:35]])
    Origin_block_dict['Origin_lat'].extend([line[36:44]])
    Origin_block_dict['Origin_long'].extend([line[45:54]])
    Origin_block_dict['Origin_epicenter_flag'].extend([line[54:55]])
    Origin_block_dict['Origin_Smaj'].extend([line[55:60]])
    Origin_block_dict['Origin_Smin'].extend([line[61:66]])    
    Origin_block_dict['Origin_Az'].extend([line[67:70]])  
    Origin_block_dict['Origin_Depth'].extend([line[71:76]])
    Origin_block_dict['Depth_fixed_flag'].extend([line[76:77]])
    Origin_block_dict['Origin_depth_errors'].extend([line[78:82]])
    Origin_block_dict['Origin_Ndef'].extend([line[83:87]])
    Origin_block_dict['Origin_Nsta'].extend([line[88:92]])
    Origin_block_dict['Origin_Gap'].extend([line[93:96]])
    Origin_block_dict['Origin_mdist'].extend([line[97:103]])
    Origin_block_dict['Origin_Mdist'].extend([line[104:110]])
    Origin_block_dict['Origin_analysis_type'].extend([line[111:112]]) 
    Origin_block_dict['Origin_location_method'].extend([line[113:114]]) 
    Origin_block_dict['Origin_event_type'].extend([line[115:117]]) 
    Origin_block_dict['Origin_author'].extend([line[118:127]])     
    Origin_block_dict['Origin_ID'].extend([line[128:36]])
    
    #df_origins=pd.DataFrame.from_dict(Origin_block_dict)
    return(Origin_block_dict)


def read_magnitude_block_line(line):
    # Magnitude_block_dict = {'Magnitude_type': [], 'Magnitude_min_max_indicator': [], \
    # 'Magnitude_value': [],'Magnitude_standart_error':[], 'Magnitude_Nsta': [], \
    # 'Magnitude_Author': [],'Magnitude_OriginID': []}

    Magnitude_block_dict['Magnitude_type'].extend([line[0:5]])
    Magnitude_block_dict['Magnitude_min_max_indicator'].extend([line[5:6]])
    Magnitude_block_dict['Magnitude_value'].extend([line[6:10]])
    Magnitude_block_dict['Magnitude_standart_error'].extend([line[11:14]])
    Magnitude_block_dict['Magnitude_Nsta'].extend([line[15:19]])
    Magnitude_block_dict['Magnitude_Author'].extend([line[20:29]])
    Magnitude_block_dict['Magnitude_OriginID'].extend([line[30:38]]) 
    
    #df_magnitudes=pd.DataFrame.from_dict(Magnitude_block_dict)
    return(Magnitude_block_dict)


def read_phase_block_line(line):
    # Phase_block_dict= {'Sta': [], 'Dist': [], 'EzAz': [], 'Phase': [], 'Arrival_time': [], 'Tres': [],'Azim': [], 'AzRes': [], 'Slow': [],\
    # 'SRes': [], 'T_flag': [], 'A_flag': [], 'S_flag': [], 'SNR': [], 'Amp': [], 'Per': [], 'Pick_evaluation_mode': [],\
    # 'Pick_polarity': [], 'Pick_onset': [], 'Magnitude_type': [], 'Min_max_indicator': [], 'Magnitude_value': [], 'ArrID': []}

    Phase_block_dict['Sta'].extend([line[0:5]])
    Phase_block_dict['Dist'].extend([line[6:12]])
    Phase_block_dict['EzAz'].extend([line[13:18]])
    Phase_block_dict['Phase'].extend([line[19:27]])
    Phase_block_dict['Arrival_time'].extend([line[28:40]])
    Phase_block_dict['Tres'].extend([line[41:46]])
    Phase_block_dict['Azim'].extend([line[47:52]])
    Phase_block_dict['AzRes'].extend([line[53:58]])
    Phase_block_dict['Slow'].extend([line[59:65]])
    Phase_block_dict['SRes'].extend([line[66:72]])    
    Phase_block_dict['T_flag'].extend([line[73:74]])  
    Phase_block_dict['A_flag'].extend([line[74:75]])
    Phase_block_dict['S_flag'].extend([line[75:76]])
    Phase_block_dict['SNR'].extend([line[77:82]])
    Phase_block_dict['Amp'].extend([line[83:92]])
    Phase_block_dict['Per'].extend([line[93:98]])
    Phase_block_dict['Pick_evaluation_mode'].extend([line[99:100]])
    Phase_block_dict['Pick_polarity'].extend([line[100:101]])
    Phase_block_dict['Pick_onset'].extend([line[101:102]])
    Phase_block_dict['Magnitude_type'].extend([line[103:108]]) 
    Phase_block_dict['Min_max_indicator'].extend([line[108:109]]) 
    Phase_block_dict['Magnitude_value'].extend([line[109:113]]) 
    Phase_block_dict['ArrID'].extend([line[114:122]]) 
    #df_phases=pd.DataFrame.from_dict(Phase_block_dict)
    return(Phase_block_dict)


def find_comments(line):
    comment=line[line.find('(#')+1:line.find(')')]
    comments_list.append(comment)
    del comment
    return comments_list



def isf_read(fname):
	
	# HAD TO INCLUDE IN ORDER TO CONVERT FROM JUPYTER NOTEBOOK TO .PY FUNCTION
	# NEEDS TO BE REWRITTEN TO AVOID THE USE OF GLOBAL VARIABLES
	global Origin_block_dict, Magnitude_block_dict, Phase_block_dict
	
	# Opening file
	file1 = open(fname, 'r')

	num_lines = sum(1 for line in open(fname, 'r'))

	# print("total number of lines in " + str(fname)+ ".isf is:", num_lines)

	count = 0 #lines counter
	origin_lines_count=0
	magnitude_lines_count=0
	phase_lines_count=0
	events_num=0 # counter for the events

	# an empty dictionary where a dictionary for each event will be appended
	Events_collection= {} 

	Origin_comments_collection={}
	Magnitude_comments_collection={}
	Phase_comments_collection={}
	#a set of boolean variables to catch the beggining of different blocks beggnings

	New_event_flag=False #a flag for a new event
	Origin_block_start=False #a flag for an origin block
	Magnitude_block_start=False # a flag for a magnitude block
	Phase_block_start=False # a flag for a phase block


	for line_no, line in enumerate(file1):

	    count += 1
	    #print(line)
	    
	    if len(line)==0:
	        'line is empty'
	    else:
	   
	        if str(line[0:5])=='Event':
	        
	            Phase_block_start=False
	            phase_lines_count=0
	            events_num= events_num + 1
	            if events_num>1:
	                New_event_flag=True
	                
	            if New_event_flag:
	                # For each event a dictionary is created consisting of 
	                # Origin, Magnitude and Phase blocks dataframes and comments
	                # block in a form of list
	                Event_dict = {'Origin block':\
	                [pd.DataFrame.from_dict(Origin_block_dict)], \
	                'Magnitude block': \
	                [pd.DataFrame.from_dict(Magnitude_block_dict)], \
	                'Phase block':[pd.DataFrame.from_dict(Phase_block_dict)],\
	                'Origin_comments_collection':\
	                [pd.DataFrame.from_dict(Origin_comments_dict)],\
	                'Magnitude_comments_collection':[\
	                pd.DataFrame.from_dict(Magnitude_comments_dict)],\
	                'Phase_comments_collection':\
	                [pd.DataFrame.from_dict(Phase_comments_dict)]}
	                # Events dictionary consists of a single event dictionary 
	                # with assigned event number as key 
	                Event_to_collection = {Event_number_key_cur: Event_dict}                
	                Events_collection.update(Event_to_collection)
	                del Event_dict

	                New_event_start=False               
	            
	            Event_number_key=''.join((re.findall(r'\d+', str(line))))

	            

	            Origin_block_dict = {'Origin_Date': [], 'Origin_Time': [], \
	            'Origin_time_flag': [],'Origin_time_error': [],'Origin_RMS': \
	            [],'Origin_lat': [],'Origin_long': [],'Origin_epicenter_flag': \
	            [], 'Origin_Smaj': [], 'Origin_Smin': [], 'Origin_Az' : [], \
	            'Origin_Depth': [], 'Depth_fixed_flag': [], \
	            'Origin_depth_errors': [], 'Origin_Ndef': [],'Origin_Nsta': \
	            [],'Origin_Gap': [],'Origin_mdist': [], 'Origin_Mdist': [],\
	            'Origin_analysis_type': [], 'Origin_location_method': [], \
	            'Origin_event_type': [], 'Origin_author': [],'Origin_ID': []}

	            Magnitude_block_dict = {'Magnitude_type': [], \
	            'Magnitude_min_max_indicator': [], 'Magnitude_value': [],\
	                        'Magnitude_standart_error':[], 'Magnitude_Nsta': \
	                        [], 'Magnitude_Author': [],'Magnitude_OriginID': []} 

	            Phase_block_dict= {'Sta': [], 'Dist': [], 'EzAz': [], 'Phase': \
	            [], 'Arrival_time': [], 'Tres': [],'Azim': [], 'AzRes': [], \
	            'Slow': [],'SRes': [], 'T_flag': [], 'A_flag': [], 'S_flag': \
	            [], 'SNR': [], 'Amp': [], 'Per': [], 'Pick_evaluation_mode': \
	            [],'Pick_polarity': [], 'Pick_onset': [], 'Magnitude_type': \
	            [], 'Min_max_indicator': [], 'Magnitude_value': [], 'ArrID': []}

	            Origin_comments_dict={'Str_num':[],'Comment':[]}
	            Magnitude_comments_dict={'Str_num':[],'Comment':[]}
	            Phase_comments_dict={'Str_num':[],'Comment':[]}
	            
	            #comments_list =[]
	            
	            continue 
	            
	        #work around to catch last event in the file
	        if count==num_lines:
	            #For each event a dictionary is created consisting of Origin, Magnitude and Phase blocks dataframes and commets block in a form of list
	            Event_dict = {'Event number':[Event_number_key_cur],'Origin block':[pd.DataFrame.from_dict(Origin_block_dict)], 'Magnitude block': [pd.DataFrame.from_dict(Magnitude_block_dict)], 'Phase block':[pd.DataFrame.from_dict(Phase_block_dict)], 'Origin_comments_collection':[pd.DataFrame.from_dict(Origin_comments_dict)],\
	                             'Magnitude_comments_collection':[pd.DataFrame.from_dict(Magnitude_comments_dict)],'Phase_comments_collection':[pd.DataFrame.from_dict(Phase_comments_dict)]}
	            #Events dictionary consists of a single event dictionary with assigned event number as key
	            Event_to_collection = {Event_number_key_cur: Event_dict}
	            Events_collection.update(Event_to_collection)
	            del Event_dict
	            New_event_start=False
	        
	        
	        if str(line[3:7])=='Date':
	            Origin_block_start = True
	            Magnitude_block_start=False
	            continue
	            
	        
	        if str(line[0:9])=='Magnitude':
	            Origin_block_start = False
	            origin_lines_count=0
	            Magnitude_block_start=True
	            continue


	        if str(line[0:3])=='Sta':
	            Origin_block_start = False
	            Magnitude_block_start=False
	            magnitude_lines_count=0
	            Phase_block_start=True
	            continue
	            
	            
	        if Origin_block_start:

	            if len(line.strip())==0:
	                continue
	                
	            Event_number_key_cur=Event_number_key
	            
	            if str(line[0:2])=='(#':
	                comment=line[line.find('(#')+1:line.find(')')]
	                Origin_comments_dict['Str_num'].append(origin_lines_count-1)
	                Origin_comments_dict['Comment'].append(comment)
	                
	                #Comment_to_collection = {origin_lines_count: comment}
	                #Origin_comments_collection.update(Comment_to_collection)
	                #origin_comments_list.append(comment)
	                del comment
	                continue
	            elif str(line[0:2])=='(+':
	                comment=line[line.find('(+')+1:line.find(')')]
	                Origin_comments_dict['Str_num'].append(origin_lines_count-1)
	                Origin_comments_dict['Comment'].append(comment)
	                #origin_comments_list.append(comment)
	                del comment
	                continue
	                
	            else:   
	                cur_Origin_dict=read_origin_block_line(line)
	                Origin_block_dict.update(cur_Origin_dict)
	                origin_lines_count+= 1
	                del cur_Origin_dict
	        
	        if Magnitude_block_start:

	            if len(line.strip())==0:
	                continue
	                
	            if str(line[0:2])=='(#':
	                comment=line[line.find('(#')+1:line.find(')')]
	                Magnitude_comments_dict['Str_num'].append(origin_lines_count-1)
	                Magnitude_comments_dict['Comment'].append(comment)

	                #magnitude_comments_list.append(comment)
	                del comment
	                continue
	            elif str(line[0:2])=='(+':
	                comment=line[line.find('(+')+1:line.find(')')]
	                Magnitude_comments_dict['Str_num'].append(origin_lines_count-1)
	                Magnitude_comments_dict['Comment'].append(comment)
	                #magnitude_comments_list.append(comment)
	                del comment
	                continue
	                
	            else:
	                cur_Magnitude_dict=read_magnitude_block_line(line)
	                Magnitude_block_dict.update(cur_Magnitude_dict)
	                magnitude_lines_count+= 1
	                del cur_Magnitude_dict

	            
	        if Phase_block_start:

	            if len(line.strip())==0:
	                continue
	                
	            if str(line[0:2])=='(#':
	                comment=line[line.find('(#')+1:line.find(')')]
	                Phase_comments_dict['Str_num'].append(origin_lines_count-1)
	                Phase_comments_dict['Comment'].append(comment)
	                #phase_comments_list.append(comment)
	                del comment
	                continue
	            elif str(line[0:2])=='(+':
	                comment=line[line.find('(+')+1:line.find(')')]
	                Phase_comments_dict['Str_num'].append(origin_lines_count-1)
	                Phase_comments_dict['Comment'].append(comment)
	                #phase_comments_list.append(comment)
	                del comment
	                continue
	                
	            else:
	                cur_Phase_dict=read_phase_block_line(line)
	                Phase_block_dict.update(cur_Phase_dict)
	                phase_lines_count+= 1
	                del cur_Phase_dict
	        
	        
	  
	# Closing files
	file1.close()

	return Events_collection

